

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>


void summarize_opt1(int N)
{
  int i, n, my_rank, my_N;
  int sum=0, tmp_sum;
  
  MPI_Comm_size(MPI_COMM_WORLD, &n);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

  my_N = N/n + (N%n>my_rank);   // numbers to summarize locally

  for(i = 0; i < my_N; i++) {
    sum+=(rand()%N);
  }

  printf("Sum at process %d: %d\n", my_rank, sum);

  if(my_rank==0) {   // master: receive sums
    for(i = 1; i < n; i++) {
      MPI_Recv(&tmp_sum, 1, MPI_INT, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
      sum+=tmp_sum;
    }
    printf("Sum (1): %d\n", sum);
  } else {          // slave: send sums
    MPI_Send(&sum, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
  }  
}

void summarize_opt2(int N) {
  int i, n, my_rank, my_N;
  int sum;
  int *recv_sum = NULL;
  
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &n);
  
  my_N = N/n + (N%n>my_rank);   // numbers to summarize locally

  for(i = 0; i < my_N; i++) {
    sum+=(rand()%N);
  }

  printf("Sum at process %d: %d\n", my_rank, sum);

  recv_sum=NULL;

  if(my_rank==0) {
    recv_sum = malloc(n*sizeof(int));
    if(recv_sum==NULL) {
      fprintf(stderr, "Error malloc.\n");
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }
  }

  // Use MPI's gather function to receive all numbers
  MPI_Gather(&sum, 1, MPI_INT, recv_sum, 1, MPI_INT, 0, MPI_COMM_WORLD);

  sum = 0;
  if(my_rank==0) {
    for(i = 0; i < n; i++) {
      printf("%d \n", recv_sum[i]);
      sum+=recv_sum[i];
    }
    printf("Sum (2): %d\n", sum);
  }
  free(recv_sum);
  
}

void summarize_opt3(int N) {
  int i, n, my_rank, my_N;
  int sum, tmp_sum;
  
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &n);
  
  my_N = N/n + (N%n>my_rank);   // numbers to summarize
  sum = 0;
  tmp_sum = 0;

  for(i = 0; i < my_N; i++) {
    sum+=(rand()%N);
  }

  printf("Sum at process %d: %d\n", my_rank, sum);
  
  // Use reduce to add numbers directly
  MPI_Reduce(&sum, &tmp_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
  if(my_rank==0)
    printf("Sum (3): %d\n", tmp_sum);
}

int main(int argc, char *argv[]) {
  int N, s;

  if(argc < 2) {
    fprintf(stderr, "Give N = number to summarize as argument.\n");
    return -1;
  }

  N = atoi(argv[1]);
  s = MPI_Init(&argc, &argv);

  if(s!=MPI_SUCCESS) {
    fprintf(stderr, "Error MPI_Init\n");
    return -1;
  }

  summarize_opt1(N);
  summarize_opt2(N);
  summarize_opt3(N);

  MPI_Finalize();

  return 0;
}
