#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

void
ColMatrixVectorMultiply(int n, double *a, double *b, double *x,
			MPI_Comm comm)
{
  int i, j;
  int nlocal;
  double *px;
  /* double *fx; */
  int npes, myrank;
  /* MPI_Status status; */

  /* Get identity and size information from the communicator */
  MPI_Comm_size(comm, &npes);
  MPI_Comm_rank(comm, &myrank);

  nlocal = n/npes;

  /* Allocate memory for arrays storing intermediate results. */
  px = (double *)malloc(n*sizeof(double));
  /* fx = (double *)malloc(n*sizeof(double)); */

  /* Compute the partial-dot products that correspond to the local columns of A. */
  for (i=0; i<n; i++) {
    px[i] = 0.0;
    for (j=0; j<nlocal; j++)
      px[i] += a[i*nlocal+j]*b[j];
  }

  for (i=0; i<npes; i++)
    MPI_Reduce(&px[i*nlocal], x, nlocal, MPI_DOUBLE, MPI_SUM, i, comm);

  free(px); /* free(fx); */
}

int main (int argc, char * argv[]) {
  int i, j;

  int npes, myrank;

  int n;
  int nlocal;

  double *x;
  double *b;
  double *a;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  if (myrank == 0) {
    if (argc < 2) {
      fprintf(stderr, "usage: %s n\n", argv[0]);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    n = atoi(argv[1]);
  }

  MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  if (n % npes) {
    if (myrank == 0)
      fprintf(stderr, "%s: value %d doesn't divide number of processes %d\n",
	      argv[0], n, npes);

    MPI_Finalize();
    return 1;
  }

  nlocal = n/npes;
  b = (double *)malloc(nlocal*sizeof(double));

  for (i=0; i<nlocal; i++)
    b[i] = (double)(myrank*nlocal+i)/n;

  a = (double *)malloc(nlocal*n*sizeof(double));

  for (i=0; i<nlocal; i++)
    for (j=0; j<n; j++)
      a[i*n+j] = (double)(myrank*nlocal*n+i+j)/(n*n);

  x = (double *)malloc(nlocal*sizeof(double));

  ColMatrixVectorMultiply(n, a, b, x, MPI_COMM_WORLD);

  free(x);
  free(a);
  free(b);

  MPI_Finalize();
  return 0;
}
