
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <mpi.h>

void ColMatrixVectorMultiply(int n, double *a, double *b, double *x, MPI_Comm comm) {
  int i, j;
  int nlocal;
  double *px, *fx;
  int npes, myrank;
  MPI_Status status;

  /* Get identity and size information from the communicator */
  MPI_Comm_size(comm, &npes);
  MPI_Comm_rank(comm, &myrank);
  
  nlocal = n/npes;
  
  /* Allocate memory for arrays storing intermediate results */
  px = (double *)malloc(n*sizeof(double));
  fx = (double *)malloc(n*sizeof(double));

  /* Compute the partial-dot products that correspond to the local columns of A */
  for(i = 0; i < n; i++) {
    px[i] = 0.0;
    for(j = 0; j < nlocal; j++) 
      px[i] += a[i*nlocal+j]*b[j];
    
  }

  MPI_Allreduce(px, fx, n, MPI_DOUBLE, MPI_SUM, comm);

  memcpy(&x[0], &fx[myrank*nlocal], nlocal);
  
  for(i = 0; i < n; i++) {
    printf("%d %f \n", myrank, x[i]);
  }
  
  free(px);
  free(fx);
}

int main(int argc, char *argv[]) {
  int i, j, n, nlocal;
  int myrank, npes;
  double *a, *b, *x;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  if (myrank == 0) {
    if (argc < 2) {
      fprintf(stderr, "usage: %s n\n", argv[0]);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    n = atoi(argv[1]);

    if(n%npes!=0) {
      fprintf(stderr, "n must be a multiplum of number of processes.\n");
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }
  }

  MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  nlocal = n/npes;

  a = malloc(n*nlocal*sizeof(double));
  b = malloc(nlocal*sizeof(double));
  x = malloc(nlocal*sizeof(double));
  
  for(i = 0; i < nlocal; i++)
    b[i] = 1.0;

  for(i = 0; i < nlocal; i++)
    for(j = 0; j < n; j++)
      a[i*n+j] = 1.0;

  ColMatrixVectorMultiply(n, a, b, x, MPI_COMM_WORLD);
  
  free(a);
  free(b);
  free(x);

  MPI_Finalize();
  return 0;
}
